@@ -1,11 +1,12 @@
 # -*- test-case-name: twisted.test.test_failure -*-
 # See also test suite twisted.test.test_pbfailure
 
-# Copyright (c) 2001-2004 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2010 Twisted Matrix Laboratories.
 # See LICENSE for details.
 
 
-"""Asynchronous-friendly error mechanism.
+"""
+Asynchronous-friendly error mechanism.
 
 See L{Failure}.
 """
@@ -13,10 +14,11 @@ See L{Failure}.
 # System Imports
 import sys
 import linecache
-import string
 import inspect
+import opcode
 from cStringIO import StringIO
-import types
+
+from twisted.python import reflect
 
 count = 0
 traceupLength = 4
@@ -46,7 +48,7 @@ def format_frames(frames, write, detail="default"):
     elif detail == "default":
         for method, filename, lineno, localVars, globalVars in frames:
             w( '  File "%s", line %s, in %s\n' % (filename, lineno, method))
-            w( '    %s\n' % string.strip(linecache.getline(filename, lineno)))
+            w( '    %s\n' % linecache.getline(filename, lineno).strip())
     elif detail == "verbose":
         for method, filename, lineno, localVars, globalVars in frames:
             w("%s:%d: %s(...)\n" % (filename, lineno, method))
@@ -60,7 +62,7 @@ def format_frames(frames, write, detail="default"):
 
 # slyphon: i have a need to check for this value in trial
 #          so I made it a module-level constant
-EXCEPTION_CAUGHT_HERE = "--- <exception caught here> ---" 
+EXCEPTION_CAUGHT_HERE = "--- <exception caught here> ---"
 
 
 
@@ -71,9 +73,65 @@ class NoCurrentExceptionError(Exception):
     """
 
 
+class _Traceback(object):
+    """
+    Fake traceback object which can be passed to functions in the standard
+    library L{traceback} module.
+    """
+
+    def __init__(self, frames):
+        """
+        Construct a fake traceback object using a list of frames. Note that
+        although frames generally include locals and globals, this information
+        is not kept by this object, since locals and globals are not used in
+        standard tracebacks.
+
+        @param frames: [(methodname, filename, lineno, locals, globals), ...]
+        """
+        assert len(frames) > 0, "Must pass some frames"
+        head, frames = frames[0], frames[1:]
+        name, filename, lineno, localz, globalz = head
+        self.tb_frame = _Frame(name, filename)
+        self.tb_lineno = lineno
+        if len(frames) == 0:
+            self.tb_next = None
+        else:
+            self.tb_next = _Traceback(frames)
+
+
+class _Frame(object):
+    """
+    A fake frame object, used by L{_Traceback}.
+
+    @ivar f_code: fake L{code<types.CodeType>} object
+    @ivar f_globals: fake f_globals dictionary (usually empty)
+    @ivar f_locals: fake f_locals dictionary (usually empty)
+    """
+
+    def __init__(self, name, filename):
+        """
+        @param name: method/function name for this frame.
+        @type name: C{str}
+        @param filename: filename for this frame.
+        @type name: C{str}
+        """
+        self.f_code = _Code(name, filename)
+        self.f_globals = {}
+        self.f_locals = {}
+
+
+class _Code(object):
+    """
+    A fake code object, used by L{_Traceback} via L{_Frame}.
+    """
+    def __init__(self, name, filename):
+        self.co_name = name
+        self.co_filename = filename
+
 
 class Failure:
-    """A basic abstraction for an error that has occurred.
+    """
+    A basic abstraction for an error that has occurred.
 
     This is necessary because Python's built-in error mechanisms are
     inconvenient for asynchronous communication.
@@ -85,13 +143,30 @@ class Failure:
     pickled = 0
     stack = None
 
-    def __init__(self, exc_value=None, exc_type=None, exc_tb=None):
-        """Initialize me with an explanation of the error.
+    # The opcode of "yield" in Python bytecode. We need this in _findFailure in
+    # order to identify whether an exception was thrown by a
+    # throwExceptionIntoGenerator.
+    _yieldOpcode = chr(opcode.opmap["YIELD_VALUE"])
 
-        By default, this will use the current X{exception}
+    def __init__(self, exc_value=None, exc_type=None, exc_tb=None):
+        """
+        Initialize me with an explanation of the error.
+
+        By default, this will use the current C{exception}
         (L{sys.exc_info}()).  However, if you want to specify a
         particular kind of failure, you can pass an exception as an
         argument.
+
+        If no C{exc_value} is passed, then an "original" C{Failure} will
+        be searched for. If the current exception handler that this
+        C{Failure} is being constructed in is handling an exception
+        raised by L{raiseException}, then this C{Failure} will act like
+        the original C{Failure}.
+
+        For C{exc_tb} only L{traceback} instances or C{None} are allowed.
+        If C{None} is supplied for C{exc_value}, the value of C{exc_tb} is
+        ignored, otherwise if C{exc_tb} is C{None}, it will be found from
+        execution context (ie, L{sys.exc_info}).
         """
         global count
         count = count + 1
@@ -99,9 +174,7 @@ class Failure:
         self.type = self.value = tb = None
 
         #strings Exceptions/Failures are bad, mmkay?
-        if ((isinstance(exc_value, types.StringType) or
-             isinstance(exc_value, types.UnicodeType))
-            and exc_type is None):
+        if isinstance(exc_value, (str, unicode)) and exc_type is None:
             import warnings
             warnings.warn(
                 "Don't pass strings (like %r) to failure.Failure (replacing with a DefaultException)." %
@@ -109,6 +182,10 @@ class Failure:
             exc_value = DefaultException(exc_value)
 
         stackOffset = 0
+
+        if exc_value is None:
+            exc_value = self._findFailure()
+
         if exc_value is None:
             self.type, self.value, tb = sys.exc_info()
             if self.type is None:
@@ -150,7 +227,7 @@ class Failure:
             # and if we were passed a plain exception with no
             # traceback, it's not useful anyway
             f = stackOffset = None
-        
+
         while stackOffset and f:
             # This excludes this Failure.__init__ frame from the
             # stack, leaving it to start with our caller instead.
@@ -193,7 +270,7 @@ class Failure:
             for d in globalz, localz:
                 if d.has_key("__builtins__"):
                     del d["__builtins__"]
-            
+
             frames.append([
                 f.f_code.co_name,
                 f.f_code.co_filename,
@@ -216,15 +293,15 @@ class Failure:
         automatically re-raised if it is not a type that you expect.
 
         The reason for having this particular API is because it's very useful
-        in Deferred errback chains:
+        in Deferred errback chains::
 
-        | def _ebFoo(self, failure):
-        |     r = failure.trap(Spam, Eggs)
-        |     print 'The Failure is due to either Spam or Eggs!'
-        |     if r == Spam:
-        |         print 'Spam did it!'
-        |     elif r == Eggs:
-        |         print 'Eggs did it!'
+            def _ebFoo(self, failure):
+                r = failure.trap(Spam, Eggs)
+                print 'The Failure is due to either Spam or Eggs!'
+                if r == Spam:
+                    print 'Spam did it!'
+                elif r == Eggs:
+                    print 'Eggs did it!'
 
         If the failure is not a Spam or an Eggs, then the Failure
         will be 'passed on' to the next errback.
@@ -251,6 +328,7 @@ class Failure:
                 return error
         return None
 
+
     def raiseException(self):
         """
         raise the original exception, preserving traceback
@@ -259,6 +337,76 @@ class Failure:
         raise self.type, self.value, self.tb
 
 
+    def throwExceptionIntoGenerator(self, g):
+        """
+        Throw the original exception into the given generator,
+        preserving traceback information if available.
+
+        @return: The next value yielded from the generator.
+        @raise StopIteration: If there are no more values in the generator.
+        @raise anything else: Anything that the generator raises.
+        """
+        return g.throw(self.type, self.value, self.tb)
+
+
+    def _findFailure(cls):
+        """
+        Find the failure that represents the exception currently in context.
+        """
+        tb = sys.exc_info()[-1]
+        if not tb:
+            return
+
+        secondLastTb = None
+        lastTb = tb
+        while lastTb.tb_next:
+            secondLastTb = lastTb
+            lastTb = lastTb.tb_next
+
+        lastFrame = lastTb.tb_frame
+
+        # NOTE: f_locals.get('self') is used rather than
+        # f_locals['self'] because psyco frames do not contain
+        # anything in their locals() dicts.  psyco makes debugging
+        # difficult anyhow, so losing the Failure objects (and thus
+        # the tracebacks) here when it is used is not that big a deal.
+
+        # handle raiseException-originated exceptions
+        if lastFrame.f_code is cls.raiseException.func_code:
+            return lastFrame.f_locals.get('self')
+
+        # handle throwExceptionIntoGenerator-originated exceptions
+        # this is tricky, and differs if the exception was caught
+        # inside the generator, or above it:
+
+        # it is only really originating from
+        # throwExceptionIntoGenerator if the bottom of the traceback
+        # is a yield.
+        # Pyrex and Cython extensions create traceback frames
+        # with no co_code, but they can't yield so we know it's okay to just return here.
+        if ((not lastFrame.f_code.co_code) or
+            lastFrame.f_code.co_code[lastTb.tb_lasti] != cls._yieldOpcode):
+            return
+
+        # if the exception was caught above the generator.throw
+        # (outside the generator), it will appear in the tb (as the
+        # second last item):
+        if secondLastTb:
+            frame = secondLastTb.tb_frame
+            if frame.f_code is cls.throwExceptionIntoGenerator.func_code:
+                return frame.f_locals.get('self')
+
+        # if the exception was caught below the generator.throw
+        # (inside the generator), it will appear in the frames' linked
+        # list, above the top-level traceback item (which must be the
+        # generator frame itself, thus its caller is
+        # throwExceptionIntoGenerator).
+        frame = tb.tb_frame.f_back
+        if frame and frame.f_code is cls.throwExceptionIntoGenerator.func_code:
+            return frame.f_locals.get('self')
+
+    _findFailure = classmethod(_findFailure)
+
     def __repr__(self):
         return "<%s %s>" % (self.__class__, self.type)
 
@@ -271,7 +419,7 @@ class Failure:
         if self.pickled:
             return self.__dict__
         c = self.__dict__.copy()
-        
+
         c['frames'] = [
             [
                 v[0], v[1], v[2],
@@ -302,6 +450,23 @@ class Failure:
         """
         self.__dict__ = self.__getstate__()
 
+    def getTracebackObject(self):
+        """
+        Get an object that represents this Failure's stack that can be passed
+        to traceback.extract_tb.
+
+        If the original traceback object is still present, return that. If this
+        traceback object has been lost but we still have the information,
+        return a fake traceback object (see L{_Traceback}). If there is no
+        traceback information at all, return None.
+        """
+        if self.tb is not None:
+            return self.tb
+        elif len(self.frames) > 0:
+            return _Traceback(self.frames)
+        else:
+            return None
+
     def getErrorMessage(self):
         """Get a string of the exception which caused this Failure."""
         if isinstance(self.value, Failure):
@@ -318,10 +483,24 @@ class Failure:
         self.printTraceback(file=io, elideFrameworkCode=elideFrameworkCode, detail=detail)
         return io.getvalue()
 
-    def printTraceback(self, file=None, elideFrameworkCode=0, detail='default'):
-        """Emulate Python's standard error reporting mechanism.
+
+    def printTraceback(self, file=None, elideFrameworkCode=False, detail='default'):
         """
-        if file is None: file = log.logerr
+        Emulate Python's standard error reporting mechanism.
+
+        @param file: If specified, a file-like object to which to write the
+            traceback.
+
+        @param elideFrameworkCode: A flag indicating whether to attempt to
+            remove uninteresting frames from within Twisted itself from the
+            output.
+
+        @param detail: A string indicating how much information to include
+            in the traceback.  Must be one of C{'brief'}, C{'default'}, or
+            C{'verbose'}.
+        """
+        if file is None:
+            file = log.logerr
         w = file.write
 
         # Preamble
@@ -334,7 +513,10 @@ class Failure:
                 hasFrames = 'Traceback'
             else:
                 hasFrames = 'Traceback (failure with no frames)'
-            w("%s: %s: %s\n" % (hasFrames, self.type, self.value))
+            w("%s: %s: %s\n" % (
+                    hasFrames,
+                    reflect.safe_str(self.type),
+                    reflect.safe_str(self.value)))
         else:
             w( 'Traceback (most recent call last):\n')
 
@@ -350,8 +532,16 @@ class Failure:
 
         # postamble, if any
         if not detail == 'brief':
-            w("%s: %s\n" % (reflect.qual(self.type),
-                            reflect.safe_str(self.value)))
+            # Unfortunately, self.type will not be a class object if this
+            # Failure was created implicitly from a string exception.
+            # qual() doesn't make any sense on a string, so check for this
+            # case here and just write out the string if that's what we
+            # have.
+            if isinstance(self.type, (str, unicode)):
+                w(self.type + "\n")
+            else:
+                w("%s: %s\n" % (reflect.qual(self.type),
+                                reflect.safe_str(self.value)))
         # chaining
         if isinstance(self.value, Failure):
             # TODO: indentation for chained failures?
@@ -360,6 +550,7 @@ class Failure:
         if detail == 'verbose':
             w('*--- End of Failure #%d ---\n' % self.count)
 
+
     def printBriefTraceback(self, file=None, elideFrameworkCode=0):
         """Print a traceback as densely as possible.
         """
@@ -376,10 +567,17 @@ DO_POST_MORTEM = True
 
 def _debuginit(self, exc_value=None, exc_type=None, exc_tb=None,
              Failure__init__=Failure.__init__.im_func):
+    """
+    Initialize failure object, possibly spawning pdb.
+    """
     if (exc_value, exc_type, exc_tb) == (None, None, None):
         exc = sys.exc_info()
         if not exc[0] == self.__class__ and DO_POST_MORTEM:
-            print "Jumping into debugger for post-mortem of exception '%s':" % exc[1]
+            try:
+                strrepr = str(exc[1])
+            except:
+                strrepr = "broken str"
+            print "Jumping into debugger for post-mortem of exception '%s':" % (strrepr,)
             import pdb
             pdb.post_mortem(exc[2])
     Failure__init__(self, exc_value, exc_type, exc_tb)
@@ -389,43 +587,6 @@ def startDebugMode():
     Failure.__init__ = _debuginit
 
 
-def visualtest():
-    def c():
-        1/0
-
-    def b():
-        c()
-
-    def a():
-        b()
-
-    def _frameworkCode(detailLevel, elideFrameworkCode):
-        try:
-            a()
-        except:
-            Failure().printTraceback(sys.stdout, detail=detailLevel, elideFrameworkCode=elideFrameworkCode)
-
-    def _moreFrameworkCode(*a, **kw):
-        return _frameworkCode(*a, **kw)
-
-    def frameworkCode(*a, **kw):
-        return _moreFrameworkCode(*a, **kw)
-
-    print 'Failure visual test case:'
-    for verbosity in ['brief', 'default', 'verbose']:
-        for truth in True, False:
-            print
-            print '----- next ----'
-            print 'detail: ',verbosity
-            print 'elideFrameworkCode:', truth
-            print
-            frameworkCode(verbosity, truth)
-
-
 # Sibling imports - at the bottom and unqualified to avoid unresolvable
 # circularity
-import reflect, log
-
-
-if __name__ == '__main__':
-    visualtest()
+import log
